@@ -7,6 +7,7 @@ import json
 import math
 import os
 import re
+import shlex
 import subprocess as sb
 import urllib.parse as urlparser
 from datetime import datetime
@@ -51,7 +52,9 @@ def _extract_values(record, paths):
         yield subrec.get(end, '')
 
 
-def decrypt_files(fnames, verbose=True, logger=None, timeout=None):
+def decrypt_files(
+    fnames, verbose=True, logger=None, timeout=None, keepfiles=False
+):
     """
     Decrypt the given file with gpg.
     This assumes that the gpg command
@@ -65,6 +68,8 @@ def decrypt_files(fnames, verbose=True, logger=None, timeout=None):
     :param logger: A logging.Logger object to print the command with
     :type timeout: Union[int, None]
     :param timeout: Number of seconds to wait for the decryption to finish
+    :type keepfiles: bool
+    :param keepfiles: Whether or not to keep the encrypted files after decryption
     :rtype: bool
     :return: Returns True if the decryption does not fail
     :raises: DecryptionError
@@ -77,15 +82,22 @@ def decrypt_files(fnames, verbose=True, logger=None, timeout=None):
     )
     if verbose and logger is not None:
         logger.info('{m}...'.format(m=cmd[:200]))
-    proc =  sb.Popen(cmd.split(), stdout=sb.PIPE, stderr=sb.PIPE)
+    proc =  sb.Popen(shlex.split(cmd), stdout=sb.PIPE, stderr=sb.PIPE)
     if proc.wait(timeout=timeout) != 0:
         err = proc.stderr.read().decode('utf8', 'ignore').strip()
+        msg = 'Failed to decrypt file names {f} with return code {rc}: {e}'
         raise DecryptionError(
-            'Failed to decrypt {f}: {e}'.format(f=' '.join(fnames), e=err)
+            msg.format(f=' '.join(fnames), e=err, rc=proc.returncode)
         )
     if verbose and logger is not None:
         for line in proc.stdout:
             logger.info(line.decode('utf8', 'ignore').strip())
+    if not keepfiles:
+        for file_ in fnames:
+            try:
+                os.remove(file_)
+            except:
+                continue
     return True
 
 
