@@ -22,8 +22,9 @@ from multiprocessing.pool import (
 from xml.etree import ElementTree
 
 from dateutil.parser import parse as parse_date
-from jinja2 import Template
+from google.cloud import bigquery
 from google.cloud.exceptions import NotFound
+from jinja2 import Template
 
 from simeon.download import utilities as downutils
 from simeon.exceptions import (
@@ -86,7 +87,7 @@ BQ2PY_TYPES = {
 BQ_DDL = """#standardSQL
 CREATE OR REPLACE TABLE {table} {cols}
 OPTIONS (
-    description = "{description}"
+    description = '''{description}'''
 ) AS
 {query}"""
 USER_INFO_COLS = OrderedDict([
@@ -186,6 +187,24 @@ def _report_pool_init(proj, safile=None):
         )
 
 
+def _get_schema_dict(sdir, table):
+    """
+    Find a matching JSON schema file in the directory sdir
+    for the given table
+    """
+    targets = ('schema_{t}.json', '{t}.json')
+    for target in targets:
+        schema_file = os.path.join(sdir, target.format(t=table))
+        if os.path.exists(schema_file):
+            with open(schema_file) as fh:
+                return json.load(fh)
+    msg = (
+        'A schema file for table {t} could not be found in the given '
+        'directory {d}'
+    )
+    raise MissingSchemaException(msg.format(t=table, d=sdir))
+
+
 def wait_for_bq_jobs(job_list):
     """
     Given a list of BigQuery load or query jobs,
@@ -228,7 +247,12 @@ def wait_for_bq_job_ids(job_list, client):
                 try:
                     rjob = client.get_job(job)
                     if rjob.state == 'DONE':
-                        out[job] = (rjob.errors or {})
+                        src = ', '.join(getattr(rjob, 'source_uris', []) or [])
+                        err = (rjob.errors or [])
+                        for e in err:
+                            if e:
+                                e['source'] = src
+                        out[job] = err
                 except NotFound:
                     msg = '{id} is not a valid BigQuery job ID'.format(id=job)
                     raise LoadJobException(msg) from None
@@ -253,6 +277,8 @@ def check_record_schema(record, schema, coerce=True, nullify=False):
     :returns: Modifies the record if needed
     :raises: SchemaMismatchException
     """
+    # Make a local copy
+    bq2py_types = BQ2PY_TYPES.copy()
     for field in schema:
         if field.get('field_type') != 'RECORD':
             if field.get('name') not in record and nullify:
@@ -265,7 +291,7 @@ def check_record_schema(record, schema, coerce=True, nullify=False):
                 record[field.get('name')] = None
             elif field.get('name') in record and coerce:
                 val = record[field.get('name')]
-                func = BQ2PY_TYPES.get(field.get('field_type'))
+                func = bq2py_types.get(field.get('field_type'))
                 if func and val is not None:
                     try:
                         record[field.get('name')] = func(val)
@@ -349,22 +375,15 @@ def make_user_info_combo(
     :rtype: None
     :return: Nothing, but writes the generated data to the outname argument
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     for (file_, _) in USER_INFO_COLS:
         file_ = os.path.join(dirname, file_)
         if not os.path.exists(file_):
             raise OSError(
                 '{f} does not exist in the SQL bundle.'.format(f=file_)
             )
-    schema_file = os.path.join(
-        schema_dir, 'schema_user_info_combo.json'
-    )
-    if not os.path.exists(schema_file):
-        raise MissingSchemaException(
-            'The schema file {f} does not exist. '
-            'Please provide a valid schema directory.'.format(f=schema_file)
-        )
-    with open(schema_file) as sfh:
-        schema = json.load(sfh).get('user_info_combo')
+    tbl = 'user_info_combo'
+    schema = _get_schema_dict(schema_dir, tbl)[tbl]
     users = dict()
     user_file = 'auth_user-analytics.sql'
     user_cols = USER_INFO_COLS.get((user_file, None))
@@ -434,7 +453,6 @@ def make_user_info_combo(
             drop_extra_keys(outcols, schema)
             check_record_schema(outrow, schema)
             zh.write(json.dumps(outrow) + '\n')
-        zh.flush()
 
 
 def course_from_block(block):
@@ -575,6 +593,39 @@ def _get_first_axis_meta(block, name, struct, mapping):
     return out
 
 
+def _get_video_durations(tarball):
+    """
+    Go through the .xml.tar.gz file, extract the video items in it, and make a
+    mapping between url names and the video durations in seconds
+
+    :type tarball: str
+    :param tarball: The .xml.tar.gz file from a course's SQL export
+    :rtype: dict
+    :return: Returns a mapping between video url names (hashes) and durations
+    """
+    tf = tarfile.open(tarball)
+    durations = {}
+    vids = [m for m in tf.getmembers() if not m.isdir() and 'video' in m.name]
+    for m in vids:
+        try:
+            root = ElementTree.fromstring(tf.extractfile(m).read())
+        except ElementTree.ParseError:
+            continue
+        url = root.attrib.get('url_name')
+        if not url:
+            continue
+        durations[url] = None
+        for elm in root:
+            if elm.tag == 'video_asset':
+                duration = (elm.attrib or {}).get('duration', None)
+                try:
+                    durations[url] = float(duration)
+                except (ValueError, TypeError):
+                    durations[url] = None
+                break
+    return durations
+
+
 def process_course_structure(data, start, mapping, parent=None):
     """
     The course structure data dictionary and starting point,
@@ -605,6 +656,7 @@ def process_course_structure(data, start, mapping, parent=None):
     targets = (
         ('name', 'display_name'), ('gformat', 'format'),
         ('due', 'due'), ('start', 'start'), ('graded', 'graded'),
+        ('visible_to_staff_only', 'visible_to_staff_only'),
     )
     for (key, target) in targets:
         item[key] = _get_first_axis_meta(
@@ -657,6 +709,7 @@ def make_course_axis(
     :rtype: None
     :return: Nothing, but writes the generated data to the outname argument
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     fname = os.path.join(dirname, 'course_structure-analytics.json')
     bundle = os.path.join(dirname, 'course-analytics.xml.tar.gz')
     for file_ in (fname, bundle):
@@ -665,6 +718,7 @@ def make_course_axis(
                 '{f} does not exist in the SQL bundle.'.format(f=file_)
             )
     itypes = _get_itypes(bundle)
+    durations = _get_video_durations(tarball=bundle)
     with open(fname) as fh:
         structure: dict = json.load(fh)
     # Find the course object (i.e. root object)
@@ -704,13 +758,13 @@ def make_course_axis(
             record['data']['itype'] = itypes.get(
                 record.get('module_id', '').split('/')[-1]
             )
+            record['data']['duration'] = durations.get(record['url_name'])
             if record['gformat']:
                 if not record.get('due'):
                     record['due'] = root_val.get('end')
                 if not record.get('start'):
                     record['start'] = root_val.get('start')
             zh.write(json.dumps(record) + '\n')
-        zh.flush()
 
 
 def make_grades_persistent(
@@ -730,33 +784,25 @@ def make_grades_persistent(
     :rtype: None
     :return: Nothing, but writes the generated data to the target files
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     infiles = dict([
         (
             'grades_persistentcoursegrade-analytics.sql',
-            (first_outname, 'schema_grades_persistent.json'),
+            (first_outname, 'grades_persistent'),
         ),
         (
             'grades_persistentsubsectiongrade-analytics.sql',
-            (second_outname, 'schema_grades_persistent_subsection.json'),
+            (second_outname, 'grades_persistent_subsection'),
         )
     ])
-    for file_, (outname, schema_file) in infiles.items():
+    for file_, (outname, tbl) in infiles.items():
         file_ = os.path.join(dirname, file_)
         outname = os.path.join(dirname, outname)
         if not os.path.exists(file_):
             raise OSError(
                 '{f} does not exist in the SQL bundle.'.format(f=file_)
             )
-        fschema_file = os.path.join(schema_dir, schema_file)
-        if not os.path.exists(fschema_file):
-            raise MissingSchemaException(
-                'The given schema file {f} does not exist. '
-                'Please provide a valid schema directory.'.format(f=fschema_file)
-            )
-        with open(fschema_file) as sfh:
-            sname, _ = os.path.splitext(schema_file)
-            sname = sname.replace('schema_', '')
-            schema = json.load(sfh).get(sname)
+        schema = _get_schema_dict(schema_dir, tbl).get(tbl)
         with open(file_) as gh, gzip.open(outname, 'wt') as zh:
             header = [c.strip() for c in gh.readline().split('\t')]
             reader = csv.DictReader(
@@ -771,7 +817,6 @@ def make_grades_persistent(
                         record[k] = None
                 check_record_schema(record, schema)
                 zh.write(json.dumps(record) + '\n')
-            zh.flush()
 
 
 def make_grading_policy(
@@ -785,13 +830,14 @@ def make_grading_policy(
 
     :type dirname: str
     :param dirname: Name of a course's directory of SQL files
-    :type schema_dir: str
+    :type schema_dir: Union[None, str]
     :param schema_dir: Directory where schema files live
     :type outname: str
     :param outname: The filename to give it to the generated report
     :rtype: None
     :return: Nothing, but writes the generated data to the target file
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     file_ = os.path.join(dirname, 'course-analytics.xml.tar.gz')
     if not os.path.exists(file_):
         raise OSError(
@@ -837,7 +883,6 @@ def make_grading_policy(
                 zh.write(
                     json.dumps(dict((k, grader.get(k)) for k in cols)) + '\n'
                 )
-            zh.flush()
 
 
 def _extract_mongo_values(record, key, subkey):
@@ -877,6 +922,7 @@ def make_forum_table(
     :rtype: None
     :return: Nothing, but writes the generated data to the target file
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     outname = os.path.join(dirname, outname)
     file_ = os.path.join(dirname, 'forum.mongo')
     if not os.path.exists(file_):
@@ -900,16 +946,7 @@ def make_forum_table(
             'sk', 'thread_type', 'title', 'visible', 'votes'
         ),
     }
-    schema_file = os.path.join(
-        schema_dir, 'schema_forum.json'
-    )
-    if not os.path.exists(schema_file):
-        raise MissingSchemaException(
-            'The schema file {f} does not exist. '
-            'Please provide a valid schema directory.'.format(f=schema_file)
-        )
-    with open(schema_file) as sfh:
-        schema = json.load(sfh).get('forum')
+    schema = _get_schema_dict(schema_dir, 'forum')['forum']
     with open(file_) as fh, gzip.open(outname, 'wt') as zh:
         for line in fh:
             record = json.loads(line)
@@ -940,10 +977,12 @@ def make_forum_table(
                             if isinstance(record[col][k], list):
                                 record[col][k] = json.dumps(record[col][k])
             record['mongoid'] = record['_id']
+            record['course_id'] = downutils.get_sql_course_id(
+                record.get('course_id') or ''
+            )
             drop_extra_keys(record, schema)
             check_record_schema(record, schema, True)
             zh.write(json.dumps(record) + '\n')
-        zh.flush()
 
 
 def make_problem_analysis(state, **extras):
@@ -959,8 +998,8 @@ def make_problem_analysis(state, **extras):
     :rtype: dict
     :return: Return a record to be loaded in problem_analysis
     """
-    maps = state.get('correct_map', {})
-    answers = state.get('student_answers', {})
+    maps = state.get('correct_map') or {}
+    answers = state.get('student_answers') or {}
     items = []
     for k, v in maps.items():
         items.append(
@@ -1001,6 +1040,7 @@ def make_student_module(
     :rtype: None
     :return: Nothing, but writes the generated data to the target files
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     pjoin = os.path.join
     outname = pjoin(dirname, outname)
     second = pjoin(dirname, 'problem_analysis.json.gz')
@@ -1009,10 +1049,10 @@ def make_student_module(
         raise OSError(
             '{f} does not exist in the SQL bundle.'.format(f=file_)
         )
-    with open(pjoin(schema_dir, 'schema_studentmodule.json')) as sfh:
-        module_schema = json.load(sfh).get('studentmodule')
-    with open(pjoin(schema_dir, 'schema_problem_analysis.json')) as sfh:
-        problem_schema = json.load(sfh).get('problem_analysis')
+    tbl = 'studentmodule'
+    module_schema = _get_schema_dict(schema_dir, tbl).get(tbl)
+    tbl = 'problem_analysis'
+    problem_schema = _get_schema_dict(schema_dir, tbl).get(tbl)
     prob_cols = ('correct_map', 'student_answers')
     with open(file_, encoding='UTF8', errors='ignore') as fh:
         header = [c.strip() for c in fh.readline().split('\t')]
@@ -1030,7 +1070,7 @@ def make_student_module(
                         record[k] = module_from_block(v or '')
                     if (v or '').lower() == 'null':
                         record[k] = None
-                check_record_schema(record, module_schema)
+                # check_record_schema(record, module_schema)
                 zh.write(json.dumps(record) + '\n')
                 try:
                     state = json.loads(
@@ -1051,8 +1091,6 @@ def make_student_module(
                 )
                 check_record_schema(panalysis, problem_schema)
                 ph.write(json.dumps(panalysis) + '\n')
-            zh.flush()
-            ph.flush()
 
 
 def _default_roles():
@@ -1080,13 +1118,14 @@ def make_roles_table(
 
     :type dirname: str
     :param dirname: Name of a course's directory of SQL files
-    :type schema_dir: str
+    :type schema_dir: Union[None, str]
     :param schema_dir: Directory where schema files live
     :type outname: str
     :param outname: The filename to give it to the generated report
     :rtype: None
     :return: Nothing, but writes the generated data to the target files
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     files = {
         'student_courseaccessrole-analytics.sql',
         'django_comment_client_role_users-analytics.sql',
@@ -1140,7 +1179,6 @@ def make_roles_table(
             if any(record.get(k) for k in staff):
                 record['roles'] = 'Staff'
             zh.write(json.dumps(record) + '\n')
-        zh.flush()
 
 
 def make_sql_tables_seq(
@@ -1168,6 +1206,7 @@ def make_sql_tables_seq(
     :rtype: bool
     :return: True if the files are generated, and False otherwise.
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     reports = (
         make_course_axis, make_forum_table, make_grades_persistent,
         make_grading_policy, make_roles_table,
@@ -1231,6 +1270,7 @@ def make_sql_tables_par(
     :rtype: bool
     :return: True if the files are generated, and False otherwise.
     """
+    schema_dir = schema_dir or SCHEMA_DIR
     reports = (
         make_course_axis, make_forum_table, make_grades_persistent,
         make_grading_policy, make_roles_table,
@@ -1288,6 +1328,7 @@ def make_table_from_sql(
     table, course_id, client, project, append=False,
     query_dir=QUERY_DIR, schema_dir=SCHEMA_DIR,
     wait=False, geo_table='geocode.geoip', youtube_table='videos.youtube',
+    **kwargs,
 ):
     """
     Generate a BigQuery table using the given table name,
@@ -1303,8 +1344,10 @@ def make_table_from_sql(
     :param client: An authenticated bigquery.Client object
     :type project: str
     :param project: GCP project id where the video_axis table is loaded.
-    :type query_dir: str
+    :type query_dir: Union[None, str]
     :param query_dir: Directory where query files are saved.
+    :type schema_dir: Union[None, str]
+    :param schema_dir: Directory where schema files live
     :type geo_table: str
     :param geo_table: Table name in BigQuery with geolocation data for IPs
     :type youtube_table: str
@@ -1314,6 +1357,8 @@ def make_table_from_sql(
     :rtype: Dict[str, Dict[str, str]]
     :return: Returns the errors dictionary from the LoadJob object tied to the query
     """
+    schema_dir = schema_dir or SCHEMA_DIR
+    query_dir = query_dir or QUERY_DIR
     latest_dataset = uputils.course_to_bq_dataset(
         course_id, 'sql', project
     )
@@ -1322,50 +1367,56 @@ def make_table_from_sql(
     )
     try:
         cols = []
-        for f in uputils.get_bq_schema(table, schema_dir):
+        fields, schema_desc = uputils.get_bq_schema(table, schema_dir)
+        for f in fields:
             cols.append(uputils.sqlify_bq_field(f))
         cols = ',\n'.join(cols)
     except MissingSchemaException:
         cols = ''
+        schema_desc = ''
     query, description = extract_table_query(table, query_dir)
     table = '{d}.{t}'.format(d=latest_dataset, t=table)
     if append:
-        config = uputils.make_bq_query_config(append=append)
+        config = uputils.make_bq_query_config(
+            append=True, plain=False, table=table
+        )
+        config.destination = table
     else:
         config = uputils.make_bq_query_config(plain=True)
         query = BQ_DDL.format(
             table=table,
-            description=description.strip(),
+            description=(schema_desc or description.strip()),
             query=query,
-            cols='({c})'.format(c=cols) if cols else ''
+            cols='({c})'.format(c=cols) if cols else '',
         )
     query = Template(query).render(
-        geo_table=geo_table, youtube_table=youtube_table
+        geo_table=geo_table, youtube_table=youtube_table,
+        course_id=course_id, latest_dataset=latest_dataset,
+        log_dataset=log_dataset, **kwargs
     )
     try:
         job = client.query(
             query.format(
-                latest_dataset=latest_dataset,
-                log_dataset=log_dataset,
-                geo_table=geo_table,
-                youtube_table=youtube_table,
-                course_id=course_id
+                latest_dataset=latest_dataset, log_dataset=log_dataset,
+                geo_table=geo_table, youtube_table=youtube_table,
+                course_id=course_id, **kwargs
             ),
-            job_id='{t}_{dt}'.format(
+            job_id_prefix='{t}_{dt}'.format(
                 t=table.replace('.', '_'),
-                dt=datetime.now().strftime('%Y%m%d%H%M%S')
+                dt=datetime.now().strftime('%Y%m%d%H%M%S%f')
             ),
             job_config=config,
         )
     except Exception as excp:
         errors = getattr(excp, 'errors', [None])
-        if errors:
+        if errors and errors[0]:
             msg = errors[0].get('message')
         else:
             msg = str(excp)
         raise SQLQueryException(msg)
     if wait:
-        wait_for_bq_job_ids([job.job_id], client)
+        status = wait_for_bq_job_ids([job.job_id], client)
+        return status[job.job_id]
     return job.errors or {}
 
 
@@ -1374,6 +1425,7 @@ def make_tables_from_sql(
     query_dir=QUERY_DIR, wait=False,
     geo_table='geocode.geoip', youtube_table='videos.youtube',
     parallel=False, fail_fast=False,
+    schema_dir=SCHEMA_DIR, **kwargs
 ):
     """
     This is the plural/multiple tables version of make_table_from_sql
@@ -1386,7 +1438,7 @@ def make_tables_from_sql(
     :param client: An authenticated bigquery.Client object
     :type project: str
     :param project: GCP project id where the video_axis table is loaded.
-    :type query_dir: str
+    :type query_dir: Union[None, str]
     :param query_dir: Directory where query files are saved.
     :type geo_table: str
     :param geo_table: Table name in BigQuery with geolocation data for IPs
@@ -1398,19 +1450,32 @@ def make_tables_from_sql(
     :param parallel: Whether the function is running in a process pool
     :type fail_fast: bool
     :param fail_fast: Whether to stop processing after the first error
+    :type schema_dir: Union[None, str]
+    :param schema_dir: Directory where schema files live
     :rtype: Dict[str, Dict[str, str]]
     :return: Return a dict mapping table names to their corresponding errors
     """
+    query_dir = query_dir or QUERY_DIR
+    schema_dir = schema_dir or SCHEMA_DIR
     if parallel:
         global report_bq_client
         client = report_bq_client
     out = dict()
+    dataset = uputils.course_to_bq_dataset(
+        course_id=course_id, file_type='sql', project=project
+    )
+    dataset = bigquery.Dataset.from_string(dataset)
+    dataset.description = gcp.DST_DESC.get('sql', '')
+    client.create_dataset(dataset, exists_ok=True)
     for table in tables:
         out[table] = make_table_from_sql(
             table=table, course_id=course_id, client=client, project=project,
             append=append, geo_table=geo_table, wait=wait,
             query_dir=query_dir, youtube_table=youtube_table,
+            schema_dir=schema_dir, **kwargs
         )
+        # If fail_fast is requested, and making the table
+        # returns a dictionary of errors, then return
         if fail_fast and out[table]:
             return out
     return out
@@ -1420,6 +1485,7 @@ def make_tables_from_sql_par(
     tables, courses, project, append=False, query_dir=QUERY_DIR,
     wait=False, geo_table='geocode.geoip', youtube_table='videos.youtube',
     safile=None, size=mp.cpu_count(), logger=None, fail_fast=False,
+    schema_dir=SCHEMA_DIR, **kwargs,
 ):
     """
     Parallel version of make_tables_from_sql
@@ -1448,9 +1514,13 @@ def make_tables_from_sql_par(
     :param logger: A Logger object with which to report steps carried out
     :type fail_fast: bool
     :param fail_fast: Whether to stop processing after the first error
+    :type schema_dir: Union[None, str]
+    :param schema_dir: Directory where schema files live
     :rtype: Dict[str, Dict[str, Dict[str, str]]]
     :return: A dict mapping course_ids to tables and their query errors
     """
+    query_dir = query_dir or QUERY_DIR
+    schema_dir = schema_dir or SCHEMA_DIR
     if len(courses) < size:
         size = len(courses)
     results = dict()
@@ -1465,13 +1535,15 @@ def make_tables_from_sql_par(
                         cid=course_id
                     )
                 )
+            kwds = dict(
+                tables=tables, course_id=course_id, client=None,
+                project=project, append=append, query_dir=query_dir, wait=wait,
+                geo_table=geo_table, youtube_table=youtube_table,
+                parallel=True, fail_fast=fail_fast, schema_dir=schema_dir,
+            )
+            kwds.update(kwargs)
             async_result = pool.apply_async(
-                func=make_tables_from_sql, kwds=dict(
-                    tables=tables, course_id=course_id, client=None,
-                    project=project, append=append, query_dir=query_dir, wait=wait,
-                    geo_table=geo_table, youtube_table=youtube_table,
-                    parallel=True, fail_fast=fail_fast,
-                )
+                func=make_tables_from_sql, kwds=kwds
             )
             results[course_id] = async_result
         for course_id in results:
