@@ -52,6 +52,31 @@ UN_DATA_FILE = os.path.join(
 )
 
 
+def get_log_record(line, lcnt, fname, logger):
+    """
+    Extract a record from a tracklog using the given string and line count
+
+    :type line: Union[str, bytes]
+    :param line: A line from a tracking log file
+    :type lcnt: int
+    :param lcnt: Line count of the given line in the tracking log file
+    :type fname: str
+    :param fname: Name of file being processed
+    :type logger: logging.Logger
+    :param logger: Logger object through which to output messages
+    :rtype: dict
+    :return: Deserialized data
+    """
+    errmsg = 'Line number {l} from file {f} not processed: {e}'
+    try:
+        line = line[line.find('{'):]
+        return json.loads(line)
+    except Exception as excp:
+        if logger:
+            logger.warning(errmsg.format(l=lcnt, f=fname, e=excp))
+        return {}
+
+
 def import_un_denominations(fname=None):
     """
     Import UN denominations data from GitHub if fname is
@@ -112,11 +137,14 @@ def make_geo_data(
         for ip_file in ip_files:
             if tracking_logs:
                 fh = gzip.open(ip_file, 'rt')
-                reader = map(json.loads, fh)
+                reader = map(
+                    lambda t: get_log_record(t[1], t[0], ip_file, logger),
+                    enumerate(fh, 1)
+                )
             else:
                 fh = open(ip_file)
                 reader = csv.DictReader(fh, fieldnames=['ip'])
-            seen = set()
+            # seen = set()
             line = 0
             while True:
                 line += 1
@@ -135,9 +163,11 @@ def make_geo_data(
                         print(msg, file=sys.stderr)
                     continue
                 ip_address = rec.get('ip')
-                if not ip_address or ip_address in seen:
+                if not ip_address:
                     continue
-                seen.add(ip_address)
+                # if not ip_address or ip_address in seen:
+                #     continue
+                # seen.add(ip_address)
                 try:
                     info = db.city(ip_address)
                     un_info = un_data.get(info.country.iso_code, {})
@@ -315,6 +345,10 @@ def main():
         )
         sys.exit(1)
     if args.command == 'extract':
+        if args.logger:
+            args.logger.info(
+                'Generating geolocation data from the IPs in the given files'
+            )
         files = []
         for file_ in args.ip_files:
             if '*' in file_:
